(* More flexible implementations *)

val A = P.Atom("A");
val B = P.Atom("B");
val C = P.Atom("C");

local
  structure Focus = Focus3
  open P
in
  fun entails (A1, A2) = Focus.prove (4, C.Decl (C.Null, ("u", A1)), A2)
  fun prove (A) = Focus.prove (4, C.Null, A)
  fun prove' n A = Focus.prove (n, C.Null, A)

  (* From Exercise 2.8 *)
  val A1a = Implies (A, Implies (B, C))
  val A1b = Implies (And (A, B), C)
  val A2a = Implies (A, And (B, C))
  val A2b = And (Implies (A, B), Implies (A, C))
  val A3a = Implies (A, Or (B, C))
  val A3b = Or (Implies (A, B), Implies (A, C))
  val A4a = Implies (Implies (A, B), C)
  val A4b = And (Or (A, C), Implies (B, C))
  val A5a = Implies (Or (A, B), C)
  val A5b = And (Implies (A, C), Implies (B, C))
  val A6a = And (A, Or (B, C))
  val A6b = Or (And (A, B), And (A, C))
  val A7a = Or (A, And (B, C))
  val A7b = And (Or (A, B), Or (A, C))

  (* Naming convention:
     f?? means should fail,
     s?? mean should succeed
  *)

  val s1a = entails (A1a, A1b)
  val s1b = entails (A1b, A1a)
  val s2a = entails (A2a, A2b)
  val s2b = entails (A2b, A2a)
  val f3a = entails (A3a, A3b)
  val s3b = entails (A3b, A3a)
  val f4a = entails (A4a, A4b)
  val s4b = entails (A4b, A4a)
  val s5a = entails (A5a, A5b)
  val s5b = entails (A5b, A5a)
  val s6a = entails (A6a, A6b)
  val s6b = entails (A6b, A6a)
  val s7a = entails (A7a, A7b)
  val s7b = entails (A7b, A7a)

  (* Other Examples *)

  (* Id : A => A *)
  val s01 = prove (Implies (A, A))
  (* K : A => B => A *)
  val s02 = prove (Implies (A, Implies (B, A)))
  (* S : (A => B => C) => (A => B) => (A => C) *)
  val s03 = prove (Implies (Implies (A, Implies (B, C)),
                            Implies (Implies (A, B),
                                     Implies (A, C))))

  (* (A & B => C) <=> (A => B => C) *)
  val s04 = prove (Implies (Implies (And (A, B), C),
                            Implies (A, Implies (B, C))))
  val s05 = prove (Implies (Implies (A, Implies (B, C)),
                            Implies (And (A, B), C)))

  (* (A => B & C) <=> (A => B) & (A => C) *)
  val s06 = prove (Implies (Implies (A, And (B, C)),
                            And (Implies (A, B), Implies (A, C))))
  val s07 = prove (Implies (And (Implies (A, B), Implies (A, C)),
                            Implies (A, And (B, C))))

  (* (A => T) <=> T *)
  val s08 = prove (Implies (Implies (A, True), True))
  val s09 = prove (Implies (True, Implies (A, True)))

  (* (A v B => C) <=> (A => C) & (B => C) *)
  val s10 = prove (Implies (Implies (Or (A, B), C),
                            And (Implies (A, C), Implies (B, C))))
  val s11 = prove (Implies (And (Implies (A, C),
                                 Implies (B, C)), Implies (Or (A, B), C)))

  (* (F => C) <=> T *)
  val s12 = prove (Implies (Implies (False, C), True))
  val s13 = prove (Implies (True, Implies (False, C)))

  (* (A => B v C) <= (A => B) v (A => C) *)
  (* next false *)
  val f13 = prove (Implies (Implies (A, Or (B, C)),
                            Or (Implies (A, B), Implies (A, C))))

  (* (A => B v C) =>(A => B) v (A => C) *)
  val s14= prove (Implies (Or (Implies (A, B), Implies (A, C)),
                           Implies (A, Or (B, C))))

  (* next true *)
  (* A => ((A => F) => F) *)
  val s15 = prove (Implies (A, Implies (Implies (A, False), False)))

  (* next false *)
  (* ((A => F) => F) => A *)
  val f16 = prove (Implies (Implies (Implies (A, False), False), A))

  (* true *)
  (* A & (B v C) <=> (A & B) v (A & C) *)
  val s17 = prove (Implies (And (A, Or (B, C)),
                            Or (And (A, B), And (A, C))))

  val s18 = prove (Implies (Or (And (A, B), And (A, C)),
                            And (A, Or (B, C))))

  (* A & F <=> F *)
  val s19 = prove (Implies (And (A, False), False))
  val s20 = prove (Implies (False, And (A, False)))

  (* A v (B & C) <=> (A v B) & (A v C) *)
  val s21 = prove (Implies (Or (A, And (B, C)),
                            And (Or (A, B), Or (A, C))))

  val s22 = prove (Implies (And (Or (A, B), Or (A, C)),
                            Or (A, And (B, C))))

  (* false *)
  (* Peirce's law : ((A => B) => A) => A *)
  val f23 = prove (Implies (Implies (Implies (A, B), A), A))

  (* false *)
  val f24 = prove (Or (Implies (A, False), A))

  (* false *)
  val f25 = prove (Implies (Implies (Implies (A, False), A), A))

  val s26 = prove (Implies (False, C))
  val s27 = prove (Implies (And (Implies (A, False), A), C))
  val f27_0 = prove' 0 (Implies (And (Implies (A, False), A), C))
  val s27_1 = prove' 1 (Implies (And (Implies (A, False), A), C))
  val s27_2 = prove' 2 (Implies (And (Implies (A, False), A), C))
  val s27_3 = prove' 3 (Implies (And (Implies (A, False), A), C))
  val s27_4 = prove' 4 (Implies (And (Implies (A, False), A), C))

  (* false *)
  val f26 = prove (Implies (True, False))

  val s29 = prove (Implies (A, And (A, A)))
  val s30 = prove (Implies (And (A, A), A))

  val s31 = prove (Implies (A, Or (A, A)))
  val s32 = prove (Implies (Or (A, A), A))

  (* ~(A => B) => A & ~B *)
  (* false *)
  val f33 = prove (Implies (Not (Implies (A, B)), And (A, Not (B))))

  (* A & ~B => ~(A => B) *)
  val s34 = prove (Implies (And (A, Not (B)), Not (Implies (A, B))))

  (* ~(A & B) => ~A v ~B *)
  (* false *)
  val f34 = prove (Implies (Not (And (A, B)), Or (Not (A), Not (B))))

  (* ~A v ~B => ~(A & B) *)
  val s36 = prove (Implies (Or (Not (A), Not (B)), Not (And (A, B))))

  (* A => A v B *)
  val s37 = prove (Implies (A, Or (A, B)))
  (* B => A v B *)
  val s38 = prove (Implies (B, Or (A, B)))
  (* A & B => A *)
  val s39 = prove (Implies (And (A, B), A))
  (* A & B => B *)
  val s40 = prove (Implies (And (A, B), B))

  (* A v ~A => ((A => B) => A) => A *)
  val s41 = prove (Implies (Or (A, Not A),
                            Implies (Implies (Implies (A, B), A), A)))

end;
